/**********************************************************************
Database reader functions for International Population.

Contents:
  readininit: Initialize a ReadinInfo structure.
  IPreader: module-specific function called by GSgenread to read in a
    country database (returning it in a ReadinInfo structure).
  DBconvert: converts the contents of a ReadinInfo structure returned
    by IPreader into a CountryInfo structure used by the IP module.

Originally written by Sheryl Kriss.
Rewritten by Cliff Shaffer, 5/92.
**********************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "ipincl.h"

/* label lists for the format statements */
char *Label1[] = {"TOTAL", "00-04", "05-09", "10-14", "15-19",
                  "20-24", "25-29", "30-34", "35-39", "40-44",
                  "45-49", "50-54", "55-59", "60-64", "65-69",
                  "70-74", "75-79", "80+"};

char *Label2[] = {"TOTAL", "-15", "15-19", "20-24", "25-29", "30-34",
                  "35-39", "40-44", "45-49", "50+"};

char *Label3[] = {"TOTAL", "-01", "01-04", "05-09", "10-14", "15-19",
                  "20-24", "25-29", "30-34", "35-39", "40-44",
                  "45-49", "50-54", "55-59", "60-64", "65-69",
                  "70-74", "75-79", "80+"};

/**********************************************************************
Initialize a ReadinInfo structure.
GLOBALS ACCESSED: None.
**********************************************************************/

void readininit(ReadinInfo *info)
{
  /* A copy of a ReadinInfo record used to initialize the caller's record. */
  static ReadinInfo Initrec = {"", -1, -1, -1L,
  /* 1990, 2000 values */      -1, -1, -1, -1, -1, -1,
  /* tot_pop */                {-1L, -1L, -1L, -1L, -1L, -1L, -1L, -1L, -1L,
				-1L, -1L, -1L, -1L, -1L, -1L, -1L, -1L, -1L},
  /* male_pop */               {-1L, -1L, -1L, -1L, -1L, -1L, -1L, -1L, -1L,
				-1L, -1L, -1L, -1L, -1L, -1L, -1L, -1L, -1L},
  /* fem_pop */                {-1L, -1L, -1L, -1L, -1L, -1L, -1L, -1L, -1L,
				-1L, -1L, -1L, -1L, -1L, -1L, -1L, -1L, -1L},
  /* tot_births */             {-1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
  /* male_births */            {-1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
  /* fem_births */             {-1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
  /* tot_deaths */             {-1, -1, -1, -1, -1, -1, -1, -1, -1,
                                -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
  /* male_deaths */            {-1, -1, -1, -1, -1, -1, -1, -1, -1,
                                -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
  /* fem_deaths */             {-1, -1, -1, -1, -1, -1, -1, -1, -1,
                                -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
  /* net migration */          -1};

  *info = Initrec;
}


/**********************************************************************
Initialize a ReadinInfo structure.
GLOBALS ACCESSED: None.
**********************************************************************/

void countryinit(CountryInfo *info)
{
  /* A copy of a CountryInfo record used to initialize the caller's record. */
  static CountryInfo Initrec = {"", "", -1, -1,
  /* 1990, 2000 values */       -1, -1, -1, -1, -1,
  /* male_pop */                {-1, -1, -1, -1, -1, -1, -1, -1, -1,
				 -1, -1, -1, -1, -1, -1, -1, -1, -1},
  /* fem_pop */                 {-1, -1, -1, -1, -1, -1, -1, -1, -1,
				 -1, -1, -1, -1, -1, -1, -1, -1, -1},
  /* tot_births */              {-1, -1, -1, -1, -1, -1, -1, -1, -1},
  /* tot_deaths */              {-1, -1, -1, -1, -1, -1, -1, -1, -1,
                                 -1, -1, -1, -1, -1, -1, -1, -1, -1},
  /* net migration */           -1};

  *info = Initrec;
}


/**********************************************************************
Module-specific reader function called by GSgenreader to read in field
from an IP country database.  IPreader is called by GSgenreader once for
each (applicable) keyword located in the database.  IPreader stores
the data found in a ReadinInfo structure.  Returns TRUE if processing
was completed successfully, FALSE if an error is detected.

GLOBALS ACCESSED: GSDebug
STATIC GLOBALS ACCESSED: Keywordlist, Label1, Label2, Label3.
**********************************************************************/
boolean IPreader(
  BigString buffer, /* string to process */
  ReaderStatusType status,        /* what to do */
  void *data)  /* place to put the information read */
{
  static int current_year = -1; /* needed to remember if reading 1990 or 2000 */
  FormatList format; /* Space for format keywords */
  int numkwds; /* Number of keywords found for format statement */
  int i;
  BigString string; /* Space for reading in stuff from datafile */
  long tempyear, temppop; /* Hunting for 1990 total population. */
  ReadinInfo *info = (ReadinInfo *)data; /*--- info was parm, but compiler
                                               complained; this keeps compiler
                                               quiet ---*/

  if(status == READ_INIT)
    /* First call for this database - initialize */
  {
    readininit(info);
    return(TRUE);
  }

  if(status == READ_FINAL)
  { /* last call for this database - check mandatory values */
    if (info->scale_factor != 1000)
      info->scale_factor = 1;
    if (info->total_pop_1990 == -1)
    {
      GSelog(0, "IPread: Missing 1990 total population\n");
      return(FALSE);
    }
    if (info->life_expect_1990 == -1)
    {
      GSelog (0, "IPread: Missing 1990 life expectancy\n");
      return(FALSE);
    }
    if (info->infant_deaths_1990 == -1)
    {
      GSelog (0, "IPread: Missing 1990 infant deaths per 1000\n");
      return(FALSE);
    }
    if (info->tot_fertility_1990== -1)
    {
      GSelog (0, "IPread: Missing 1990 total fertility rate\n");
      return(FALSE);
    }
    return(TRUE);
  }

  /* status == READ_PROCESS */
  if (GSstricmp (buffer, "name=") == 0)
  {
    if (!GSrdphrase(info->country_name, NAMELENGTH))
      return(FALSE);
    GSelog (0, "IPreader: Processing database for |%s|\n", info->country_name);
  }
  else if (GSstricmp(buffer, "size=") == 0)
    info->scale_factor = (int)GSrdinteger();
  else if (GSstricmp(buffer, "year=") == 0)
    current_year = (int)GSrdinteger();
  else if (GSstricmp(buffer, "total_population=") == 0)
    info->total_pop_1990 = GSrdinteger();
  else if (GSstricmp(buffer, "life_expectancy=") == 0)
  {
    if ((current_year == 1990) || (current_year == 1992))
       /* We treat everything as 1990, but our database
          is really for 1992 for most countries*/
      info->life_expect_1990 = (int)(LETFRRES * GSrdreal() + 0.5);
    else if (current_year == 2000)
      info->life_expect_2000 = (int)(LETFRRES * GSrdreal() + 0.5);
  }
  else if (GSstricmp(buffer, "infant_deaths=") == 0)
  {
    if ((current_year == 1990) || (current_year == 1992))
      info->infant_deaths_1990 = (int)(10.0 * GSrdreal() + 0.5);
    else if (current_year == 2000)
      info->infant_deaths_2000 = (int)(10.0 * GSrdreal() + 0.5);
  }
  else if (GSstricmp(buffer, "total_fertility=") == 0)
  {
    if ((current_year == 1990) || (current_year == 1992))
      info->tot_fertility_1990 = (int)(LETFRRES * GSrdreal() + 0.5);
    else if (current_year == 2000)
      info->tot_fertility_2000 = (int)(LETFRRES * GSrdreal() + 0.5);
  }
  else if (GSstricmp(buffer, "migration=") == 0)
    info->mig = GSrdinteger();
  else if (GSstricmp(buffer, "format=") == 0)
  {
    if (!GSrdformat(&numkwds, format))
      return(FALSE); /* GSrdformat should say why */

    if((numkwds == 2) &&
       (strcmp(format[0],"year") == 0) && (strcmp(format[1],"pop") == 0))
    { /* Midyear population estimates - Grab the 1990 total population
         estimate from this table. */
      if(!GSrdword(string, BIGSTRINGLEN))
        return(FALSE);
      while(strcmp(string, "endformat") != 0)
      {
        tempyear = atoi(string);
        temppop = GSrdinteger();
        if(tempyear == 1990)
          info->total_pop_1990 = temppop;
        if(!GSrdword(string, BIGSTRINGLEN))
          return(FALSE);
      }
    }

    if((numkwds == 4) && (current_year != 2000) &&
       (strcmp(format[0],"age") == 0)       &&
       (strcmp(format[1],"total_pop") == 0) &&
       (strcmp(format[2],"m_pop") == 0)     &&
       (strcmp(format[3],"f_pop") == 0))
    { /* 1990 MIDYEAR POPULATION, BY AGE AND SEX tabular data */
      for(i=0; i<18; i++)
      {
        if(!GSrdword(string, BIGSTRINGLEN))
          return(FALSE);
        if(strcmp(string, Label1[i]) != 0)
        {
          GSelog(0, "IPreader: Bad label in format: expected |%s|, got |%s|\n",
                 Label1[i], string);
          return(FALSE);
        }
        info->tot_pop[i] = GSrdinteger();
        info->male_pop[i] = GSrdinteger();
        info->fem_pop[i] = GSrdinteger();
      }
      if(!GSrdword(string, BIGSTRINGLEN))
        return(FALSE);
      if(strcmp(string, "endformat") != 0)
      {
        GSelog (0, "IPreader: Error in datafile: endformat expected.\n");
        return(FALSE);
      }
    }

    if ((numkwds == 4) && (strcmp(format[0],"age") == 0)          &&
                          (strcmp(format[1],"total_births") == 0) &&
                          (strcmp(format[2],"m_births") == 0)     &&
                          (strcmp(format[3],"f_births") == 0))
    { /* LIVE BIRTHS BY SEX, AND AGE OF MOTHER tabular data */
      for(i=0; i<10; i++)
      {
        if(!GSrdword(string, BIGSTRINGLEN))
          return(FALSE);
        if(strcmp(string, Label2[i]) != 0)
        {
          GSelog(0, "IPreader: Bad label in format: expected |%s|, got |%s|\n",
                 Label2[i], string);
          return(FALSE);
        }
        info->tot_births[i] = GSrdinteger();
        info->male_births[i] = GSrdinteger();
        info->fem_births[i] = GSrdinteger();
      }
      if(!GSrdword(string, BIGSTRINGLEN))
        return(FALSE);
      if(strcmp(string, "endformat") != 0)
      {
        GSelog (0, "IPreader: Error in datafile: endformat expected.\n");
        return(FALSE);
      }
    }

    if ((numkwds == 4) && (strcmp(format[0],"age") == 0)          &&
                          (strcmp(format[1],"total_deaths") == 0) &&
                          (strcmp(format[2],"m_deaths") == 0)     &&
                          (strcmp(format[3],"f_deaths") == 0))
    { /* DEATHS BY AGE AND SEX tabular data */
      for(i=0; i<19; i++)
      {
        if(!GSrdword(string, BIGSTRINGLEN))
          return(FALSE);
        if(strcmp(string, Label3[i]) != 0)
        {
          GSelog(0, "IPreader: Bad label in format: expected |%s|, got |%s|\n",
                 Label3[i], string);
          return(FALSE);
        }
        info->tot_deaths[i] = GSrdinteger();
        info->male_deaths[i] = GSrdinteger();
        info->fem_deaths[i] = GSrdinteger();
      }
      if(!GSrdword(string, BIGSTRINGLEN))
        return(FALSE);
      if(strcmp(string, "endformat") != 0)
      {
        GSelog (0, "IPreader: Error in datafile: endformat expected.\n");
        return(FALSE);
      }
    }
  } /* else if "format=" */
  return(TRUE);
}


/**********************************************************************
Convert the IP database info collected by the database reader to
the form used by the IP module.  This includes any necessary
approximations to make up for missing information from the original
database.  Returns TRUE if a valid country record could be created;
FALSE otherwise.
**********************************************************************/
boolean DBconvert(
  ReadinInfo *in,
  CountryInfo *out)
{
  long sum;     /* used to sum up columns of input data */
  long iscale;     /* used to scale birth rates */
  int i;
  long fem_deathrate[18];  /* female deathrates */
  int fem_life_exp;  /* Used to calculate sex-specific life */
  int raw_life_exp;  /* expectancy adjustment */

  countryinit(out);
  strcpy(out->country_name, in->country_name);
  /* These are mandatory elements. */
  /* IPreader has already checked on their existance */
  out->scale_factor = in->scale_factor;
  out->life_expectancy = in->life_expect_1990;
  out->tot_deathrate[0] = (long)in->infant_deaths_1990 * RATERES/10L;
  out->tot_fertility_rate = in->tot_fertility_1990;

  /* Year 2000 values. */
  if(in->life_expect_2000 != -1)
    out->life_expect_2000 = in->life_expect_2000;
  else
    out->life_expect_2000 = in->life_expect_1990;
  if(in->infant_deaths_2000 != -1)
    out->infant_deaths_2000 = (long)in->infant_deaths_2000 * RATERES/10L;
  else
    out->infant_deaths_2000 = (long)in->infant_deaths_1990 * RATERES/10L;
  if(in->tot_fertility_2000 != -1)
    out->tot_fertility_2000 = in->tot_fertility_2000;
  else
    out->tot_fertility_2000 = in->tot_fertility_1990;


  /* Now, fill in cohort data. */

  /* Process the cohort population totals.  */
  for(i=1; i<18; i++)
  {
    if(in->male_pop[i] != -1)
      out->male_1990_pop[i] = in->male_pop[i];
    else return(FALSE);
    if(in->fem_pop[i] != -1)
      out->fem_1990_pop[i] = in->fem_pop[i];
    else return(FALSE);
  }

  /* Since the cohort values may not sum to the total, or since the
     year for cohort data may not be 1990, these should be scaled to
     fit the 1990 total. */
  for(sum=0, i=1; i<18; i++)
    sum += out->male_1990_pop[i] + out->fem_1990_pop[i];
  if(sum != in->total_pop_1990)
  {
    if(GSDebug)
      GSelog (0, "Debug: scaling total population from %ld to %ld\n",
           sum, in->total_pop_1990);
    for(i=1; i<18; i++)
    {
      out->male_1990_pop[i] =
           RESCALE(out->male_1990_pop[i], in->total_pop_1990, sum);
      out->fem_1990_pop[i] =
           RESCALE(out->fem_1990_pop[i], in->total_pop_1990, sum);
    }
  }
  /* Now, set total to sum of rescaled cohorts */
  out->male_1990_pop[0] = 0;
  out->fem_1990_pop[0] = 0;
  for(i=1; i<18; i++)
  {
    out->male_1990_pop[0] += out->male_1990_pop[i];
    out->fem_1990_pop[0] += out->fem_1990_pop[i];
  }

  iscale = (RATERES * 1000L)/in->scale_factor;
  /* Process the cohort birth rates */
  for(i=0; i<9; i++)
    if(in->tot_births[i+1] != -1)
      /* Use cohort "10-14" population for computing "-15" rates. */
      out->tot_birthrate[i] =
           RESCALE(in->tot_births[i+1], iscale, out->fem_1990_pop[i+3]);
    else return(FALSE);
  /* Births are often under-reported, so need to scale birth rates to
     match total fertility rate. */
  set_to_tfr_target(out->tot_fertility_rate, out->tot_birthrate);

  /* Process the cohort death rates */
  /* Start at 1 because 0 is infants - which has already been set */
  for(i=1; i<18; i++)
    if(in->tot_deaths[i+1] != -1)
      out->tot_deathrate[i] = RESCALE(in->tot_deaths[i+1], iscale,
				out->male_1990_pop[i]+out->fem_1990_pop[i]);
    else return(FALSE);

  /* Calculate female deathrates to use in life expectancy adjustment */
  for(i=0; i<18; i++)
    if(in->fem_deaths[i+1] != -1)
      fem_deathrate[i] =
          RESCALE(in->fem_deaths[i+1], iscale, out->fem_1990_pop[i]);
    else
      fem_deathrate[i] = out->tot_deathrate[i];

  /* Calculate adjustment factor for male vs. female life expectancy */
  fem_life_exp = get_life_expect(fem_deathrate);
  raw_life_exp = get_life_expect(out->tot_deathrate);
  out->life_adjust = (int)IDIV(((long)fem_life_exp-(long)raw_life_exp) * 1000,
                               raw_life_exp);

  /* Deaths are often under-reported, so need to scale death rates to
     match life expectancy. */
  set_to_life_target(out->life_expectancy, out->tot_deathrate, FALSE);

  /* Init net migration */
  if (in->mig != -1)
    out->mig = in->mig;
  else
    out->mig = 0;
  return(TRUE);
}
