/**********************************************************************
International Population simulation routines.
**********************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#ifdef MSDOS
#include <alloc.h>
#endif

#include "ipincl.h"

#define SECANT(x1, y1, x2, y2, Y) (RESCALE((Y-y1), (x2-x1), (y2-y1)) + x1)
#define MAXITERATIONS 10  /* Maximum number of times to iterate
                             when adjusting life expectancy */

/*  used by dosim routine  */
static long *male, *fem;
static long *brate, *drate;

/**********************************************************************
Return the total fertility rate calculated from the given cohort
fertility rates.
**********************************************************************/
int get_TFR(long *brates)
{
  int i;
  long sum;

  for(sum=0, i=0; i<9; i++)
    sum += brates[i];
  sum *= 5;  /* multiply by number of years in each cohort
                          (assume "-15" cohort has rates for 5 years) */
  return((int)IDIV(sum, 10000)); /* 100,000/LEFTRRES*10 for children in 10ths */
}


/**********************************************************************
Return the life expectancy calculated from the given cohort death
rates.
**********************************************************************/
int get_life_expect(long *drates)
{
  long population;
  long sum;
  int cnt;
  long deaths;

  /* initialize population to an arbitrary value */
  population = 100000L;
  sum = 0;

  /* incorporate new born deaths */
  population = population - RESCALE(population, drates[0], 1000L*RATERES);
  for (cnt=0; cnt<80; cnt++)
  {
    deaths = RESCALE(population, drates[1 + cnt/5], 1000L*RATERES);
    sum += deaths * cnt + (deaths>>1);
    population -= deaths;
  }

/* if age 80 or older */
  for (cnt=80; cnt<110; cnt++)
  {
    deaths = RESCALE(population, drates[17], 1000L*RATERES);
    sum += deaths * cnt + (deaths>>1);
    population -= deaths;
  }

  /* incorporate remaining people */
  sum += population * cnt + (population>>1);
  return((int)IDIV(sum, 1000)); /* sum / (100,000 / LEFTRRES) */
}


/**********************************************************************
Set cohort fertility rates as dictated by the target value.
**********************************************************************/
void set_to_tfr_target(int target, long *brates)
{
  int i;
  int currTFR = get_TFR(brates);

  if(currTFR != target)
  {
    if(GSDebug)
      GSelog(0, "Debug: Scaling TFR from %d to %d\n", currTFR, target);
    if (currTFR != 0)
      for(i=0; i<9; i++)
      {
	brates[i] = RESCALE(brates[i], target, currTFR);
        if (brates[i] > 999*RATERES)
          brates[i] = 999*RATERES;
    }
  }
}


/**********************************************************************
Set cohort death rates as dictated by the target value.  This is
an iterative process using a life table to estimate the life expectancy.
The loop is: REPEAT estimate life expectancy, scale new values UNTIL
convergence.  Uses the secant method to speed convergence.  If (infant)
then change infant mortality rates, else do scaling without changing
infant mortality rates.
**********************************************************************/
#define SCALERES 100000L  /* Scale resolution */
void set_to_life_target(int target, long *drates, boolean infant)
{
  int iterate_cnt = 0; /* # of iterations performed to scale life expect */
  long prev, curr;
  unsigned long scale, prevscale;
  long temp;
  long holder[18]; /* copy of original deathrates */
  int startindex;    /* start calculation with or without infant rates */
  int i;

  prev = get_life_expect(drates);
  prevscale = SCALERES;
  if (abs(prev - target) >= 5)
  { /* do iterative processing until convergence */
    if(GSDebug)
      GSelog(0, "Debug: Scaling life expectancy from %ld to %d\n",
           prev, target);
    for(i=0; i<18; i++) /* save a copy of deathrates */
      holder[i] = drates[i];
    scale = IDIV(prev * SCALERES, (long)target);
    if (infant)
      startindex = 0;
    else
      startindex = 1;
    for(i=startindex; i<18; i++)
    {
      drates[i] = IDIV(holder[i] * prev, (long)target);
      if (drates[i] > 999*RATERES)
	drates[i] = 999*RATERES;
    }
    curr = get_life_expect(drates);
    iterate_cnt = 1;
    if(GSDebug)
      GSelog(0, "Debug: Iteration %d: %ld, scale %ld\n",
             iterate_cnt, curr, scale);
    while ((iterate_cnt < MAXITERATIONS) && (IDIV(curr,10) != IDIV((long)target,10)))
    {
      if ((prevscale == scale) || (curr == prev))
	break;
      temp = SECANT(prevscale, prev, scale, curr, (long)target);
      prevscale = scale;
      if (temp < 0)
	scale >>= 1;
      else
	scale = temp;
      prev = curr;
      for(i=startindex; i<18; i++)
      {
	drates[i] = RESCALE(holder[i], scale, SCALERES);
	if (drates[i] > 999*RATERES)
          drates[i] = 999*RATERES;
      }
      curr = get_life_expect(drates);
      iterate_cnt++;
      if(GSDebug)
        GSelog(0, "GSDebug: Iteration %d: %ld, scale %ld\n",
             iterate_cnt, curr, scale);
    } /* while */
    /*-------------------------------------------------------------------------
    if (abs(curr - target) >= 5)
      GSelog(0, "Warning: scaled life expectancy not accurate: %ld not %d\n",
           curr, target);
    -------------------------------------------------------------------------*/
  } /* if (abs(prev - target) >= 5) */
}


/**********************************************************************
Set cohort birth rates as dictated by the target value.
**********************************************************************/
void calc_sim_tfr_variant(IPStatetype *sptr)
{
  int currval, valdiff;
  int target;
  int yeardiff;
  int stepcnt = sptr->stepcnt;

  currval = get_TFR(sptr->birthrate[stepcnt]);
  if (currval != sptr->tfr_targ_val[stepcnt])
  { /* The values are different - do the change */
    if (sptr->year == sptr->tfr_targ_year[stepcnt])
      /* We are at the target year - must be an immediate change. */
      target = sptr->tfr_targ_val[stepcnt];
    else
    { /* We are moving toward a target year - change by a fraction */
      valdiff = sptr->tfr_targ_val[stepcnt] - currval;
      yeardiff = sptr->tfr_targ_year[stepcnt] - sptr->year;
      target = currval + IDIV(valdiff, yeardiff);
    }
    set_to_tfr_target(target, sptr->birthrate[stepcnt]);
  }
}


/**********************************************************************
Set cohort death rates as dictated by the target value.
**********************************************************************/
void calc_sim_life_variant(IPStatetype *sptr)
{
  int currval, target;
  long valdiff;
  int yeardiff;
  int stepcnt = sptr->stepcnt;
  boolean update_flag = TRUE;  /* Change inf mortality if possible */

  /* If working toward the "expected" scenario, must deal with
     infant mortality adjustment - but only if this is not an
     update caused by setting the current life expectancy */
  if (sptr->year == sptr->lifex_targ_year[stepcnt])
    sptr->infmort_targ_year[stepcnt] = -1; /* clobber this if user changed
                                              life expectancy */
  if (sptr->infmort_targ_year[stepcnt] > sptr->year)
  { /* deal with infant mortality update */
    valdiff = sptr->infmort_targ_val[stepcnt] - sptr->deathrate[stepcnt][0];
    yeardiff = sptr->infmort_targ_year[stepcnt] - sptr->year;
    sptr->deathrate[stepcnt][0] += IDIV(valdiff, yeardiff);
    update_flag = FALSE;  /* don't change this */
  }
  currval = get_life_expect(sptr->deathrate[stepcnt]);
  if (currval != sptr->lifex_targ_val[stepcnt])
  { /* The values are different - do the change */
    if (sptr->year == sptr->lifex_targ_year[stepcnt])
      /* We are at the target year - must be an immediate change. */
      target = sptr->lifex_targ_val[stepcnt];
    else
    { /* We are moving toward a target year - change by a fraction */
      valdiff = sptr->lifex_targ_val[stepcnt] - currval;
      yeardiff = sptr->lifex_targ_year[stepcnt] - sptr->year;
      target = currval + (int)IDIV(valdiff, yeardiff);
    }
    set_to_life_target(target, sptr->deathrate[stepcnt], update_flag);
  }
}


/**********************************************************************
Set net migration rate as dictated by the target value.
**********************************************************************/
void calc_sim_mig_variant(IPStatetype *sptr)
{
  long currval, valdiff;
  long target;
  int yeardiff;
  int stepcnt = sptr->stepcnt;

  currval = sptr->mig[stepcnt];
  if (currval != sptr->mig_targ_val[stepcnt])
  { /* The values are different - do the change */
    if (sptr->year == sptr->mig_targ_year[stepcnt])
      /* We are at the target year - must be an immediate change. */
      target = sptr->mig_targ_val[stepcnt];
    else
    { /* We are moving toward a target year - change by a fraction */
      valdiff = sptr->mig_targ_val[stepcnt] - currval;
      yeardiff = sptr->mig_targ_year[stepcnt] - sptr->year;
      target = currval + IDIV(valdiff, yeardiff);
    }
    sptr->mig[stepcnt] = target;
  }
}



/**********************************************************************
Copy the information of a simulation data point.
**********************************************************************/
static void copypt(IPStatetype *sptr)
{
  int stepcnt = sptr->stepcnt;
  int i;
  int prevstep = stepcnt - 1;

  for (sptr->totpop[stepcnt]=0, i=0; i<17; i++)
  { /* Cohort values might have been changed by user */
    sptr->totpop[stepcnt] += sptr->male_pop[prevstep][i];
    sptr->totpop[stepcnt] += sptr->fem_pop[prevstep][i];
  }
  sptr->tfr_targ_val[stepcnt] = sptr->tfr_targ_val[prevstep];
  sptr->tfr_targ_year[stepcnt] = sptr->tfr_targ_year[prevstep];
  sptr->lifex_targ_val[stepcnt] = sptr->lifex_targ_val[prevstep];
  sptr->lifex_targ_year[stepcnt] = sptr->lifex_targ_year[prevstep];
  sptr->infmort_targ_val[stepcnt] = sptr->infmort_targ_val[prevstep];
  sptr->infmort_targ_year[stepcnt] = sptr->infmort_targ_year[prevstep];
  sptr->mig_targ_val[stepcnt] = sptr->mig_targ_val[prevstep];
  sptr->mig_targ_year[stepcnt] = sptr->mig_targ_year[prevstep];
  for (i=0; i<17; i++)
  {
    sptr->male_pop[stepcnt][i] = sptr->male_pop[prevstep][i];
    sptr->fem_pop[stepcnt][i] = sptr->fem_pop[prevstep][i];
  }
  for (i=0; i<9; i++)
    sptr->birthrate[stepcnt][i] = sptr->birthrate[prevstep][i];
  for (i=0; i<18; i++)
    sptr->deathrate[stepcnt][i] = sptr->deathrate[prevstep][i];
  sptr->mig[stepcnt] = sptr->mig[prevstep];
}


/**********************************************************************
We treat the stated cohort birthrates, deathrates and pop totals
as samples from some curve.  Here, we break the 5 year cohorts
into 1 year cohorts, fitting the values using linear interpolation
between the sample values.  We use a kludgy method of adjusting
the totals across the cohort to keep peaks and vallys from messing
things up.
**********************************************************************/
void setup_pop(IPStatetype *sptr, long male[], long fem[])
{
  long male_center[19], fem_center[19]; /* bin centers for computation */
  long mtemp, ftemp;
  int stepcnt = sptr->stepcnt;
  int c, c5, y;

  for (c=1; c<=17; c++)
  {
    male_center[c] = sptr->male_pop[stepcnt][c-1];
    fem_center[c] = sptr->fem_pop[stepcnt][c-1];
  }
  /* continue curve at end */
  male_center[0] = 2 * male_center[1] - male_center[2];
  fem_center[0] = 2 * fem_center[1] - fem_center[2];
  male_center[18] = 2 * male_center[17] - male_center[16];
  fem_center[18] = 2 * fem_center[17] - fem_center[16];

  /* now, convert to 1 year bins */
  for(c=0; c<17; c++)
  {
    c5 = c*5;
    male[c5  ] = IDIV(8*male_center[c] + 12*male_center[c+1], 10);
    male[c5+1] = IDIV(4*male_center[c] + 16*male_center[c+1], 10);
    male[c5+2] = male_center[c+1] * 2;
    male[c5+3] = IDIV(16*male_center[c+1] + 4*male_center[c+2], 10);
    male[c5+4] = IDIV(12*male_center[c+1] + 8*male_center[c+2], 10);
    fem[c5  ] = IDIV(8*fem_center[c] + 12*fem_center[c+1], 10);
    fem[c5+1] = IDIV(4*fem_center[c] + 16*fem_center[c+1], 10);
    fem[c5+2] = fem_center[c+1] * 2;
    fem[c5+3] = IDIV(16*male_center[c+1] + 4*fem_center[c+2], 10);
    fem[c5+4] = IDIV(12*male_center[c+1] + 8*fem_center[c+2], 10);

  /* Now, rescale cohort if the total isn't right */
    mtemp = ftemp = 0;
    for (y=c5; y<c5+5; y++)
    {
      if (male[y] < 0)
	male[y] = 0;
      mtemp += male[y];
      if (fem[y] < 0)
	fem[y] = 0;
      ftemp += fem[y];
    }
    for (y=c5; y<c5+5; y++)
    {
      male[y] = RESCALE(male[y], male_center[c+1] * 10, mtemp);
      fem[y] = RESCALE(fem[y], fem_center[c+1] * 10, ftemp);
    }
  }
}


/**********************************************************************
Set up one year birthrates.
**********************************************************************/
void setup_brate(IPStatetype *sptr, long brate[], long fem[])
{
  long bcenter[11]; /* bin centers for computation */
  int stepcnt = sptr->stepcnt;
  int c, c5, y;
  unsigned long computed, expected;

  for (c=1; c<10; c++)
    bcenter[c] = sptr->birthrate[stepcnt][c-1];
  /* continue curve at end */
  bcenter[0] = 0;
  bcenter[10] = 0;

  /* now, convert to 1 year bins */
  for(c=0; c<9; c++)
  {
    c5 = c*5;
    brate[c5+10] = IDIV(4*bcenter[c] + 6*bcenter[c+1], 10);
    brate[c5+11] = IDIV(2*bcenter[c] + 8*bcenter[c+1], 10);
    brate[c5+12] = bcenter[c+1];
    brate[c5+13] = IDIV(8*bcenter[c+1] + 2*bcenter[c+2], 10);
    brate[c5+14] = IDIV(6*bcenter[c+1] + 4*bcenter[c+2], 10);

  /* Now, rescale rate if the total isn't right */
    for (computed=0, y=c5+10; y<c5+15; y++)
      computed += RESCALE(brate[y], fem[y], 100000L);
    if (computed != 0)
    {
      expected = RESCALE(bcenter[c+1], sptr->fem_pop[stepcnt][c+2], 10000);
      for (y=c5+10; y<c5+15; y++)
        brate[y] = RESCALE(brate[y], expected, computed);
    }
    else
      for (y=c5+10; y<c5+15; y++)
        brate[y] = 0;
  }
  /* finally, limit rate to be < 1 */
  for (c=0; c<COHORTS; c++)
    if (brate[c] > 999000L)
      brate[c] = 999000L;
}


/**********************************************************************
Set up one year deathrates.
**********************************************************************/
void setup_drate(IPStatetype *sptr, long drate[], long male[], long fem[])
{
  long dcenter[19]; /* bin centers for computation */
  unsigned long computed, expected;
  int c, c5, y;
  int stepcnt = sptr->stepcnt;

  for (c=1; c<18; c++)
    dcenter[c] = sptr->deathrate[stepcnt][c];
  /* continue curve at end */
  dcenter[0] = 2*dcenter[1] - dcenter[2];
  if (dcenter[0] < 0)
    dcenter[0] = 0;
  dcenter[18] = 2*dcenter[17] - dcenter[16];
  if (dcenter[18] < 0)
    dcenter[18] = 0;

  /* now, convert to 1 year bins */
  for(c=0; c<17; c++)
  {
    c5 = c * 5;
    drate[c5  ] = IDIV(4*dcenter[c] + 6*dcenter[c+1], 10);
    drate[c5+1] = IDIV(2*dcenter[c] + 8*dcenter[c+1], 10);
    drate[c5+2] = dcenter[c+1];
    drate[c5+3] = IDIV(8*dcenter[c+1] + 2*dcenter[c+2], 10);
    drate[c5+4] = IDIV(6*dcenter[c+1] + 4*dcenter[c+2], 10);

    drate[0] = 0;  /* don't hit infants twice */

  /* Now, rescale rate if the total isn't right */
    for (computed=0, y=c5; y<c5+5; y++)
      computed += RESCALE(drate[y], (male[y] + fem[y]), 100000L);
    if (computed != 0)
    {
      expected = RESCALE(dcenter[c+1],
             (sptr->male_pop[stepcnt][c] + sptr->fem_pop[stepcnt][c]), 10000);
      for (y=c5; y<c5+5; y++)
        drate[y] = RESCALE(drate[y], expected, computed);
    }
    else
      for (y=c5; y<c5+5; y++)
        drate[y] = 0;
  }
  /* finally, limit rate to be < 1 */
  for (c=0; c<COHORTS; c++)
    if (drate[c] > 999000L)
      drate[c] = 999000L;
}


/**********************************************************************
Reconvert 1 year cohorts to 5 year cohorts.
**********************************************************************/
void reconstruct_pop(IPStatetype *sptr, long *male, long *fem)
{
  int c, c5, y;
  int stepcnt = sptr->stepcnt;
  long mtemp, ftemp;

  for (c=0; c<17; c++)
  {
    c5 = c*5;
    mtemp = ftemp = 0;
    for (y=c5; y<c5+5; y++)
    {
      mtemp += male[y];
      ftemp += fem[y];
    }
    sptr->male_pop[stepcnt][c] = IDIV(mtemp, 10);
    sptr->fem_pop[stepcnt][c] = IDIV(ftemp, 10);
  }
}


boolean init_sim_arrays(void)
  {
  /*  initialize arrays for dosim  */
  if ((male = (long *)malloc (sizeof (long)*COHORTS)) == NULL)
    return (0);
  if ((fem = (long *)malloc (sizeof (long)*COHORTS)) == NULL)
    return (0);
  if ((brate = (long *)malloc (sizeof (long)*COHORTS)) == NULL)
    return (0);
  if ((drate = (long *)malloc (sizeof (long)*COHORTS)) == NULL)
    return (0);
  return (1);
  }


/**********************************************************************
Calculate the next simulation point.
**********************************************************************/
void dosim(IPStatetype *sptr)
{
  int y, i, stepcnt;
  long mig; /* Yearly net migration */
  /* adjustment factor for male/female life expectancies */
  long sexadjust = Country[Cdisp].life_adjust;
  long deathrate, deadwomen, inew_births, adj;

  copypt(sptr);
  stepcnt = sptr->stepcnt;
  mig = sptr->mig[stepcnt];
  for (y=0; y<COHORTS; y++)
    brate[y] = 0;
  /* initialize 1 year cohorts */
  setup_pop(sptr, male, fem);

  for (y=1; y<= TIME_STEP; y++)
  { /* do 5 years of simulation */
    setup_brate(sptr, brate, fem);
    setup_drate(sptr, drate, male, fem);

    /* Calculate new births. */
    for (inew_births=0, i=0; i<COHORTS; i++)
    {
      deathrate = drate[i] - IDIV(drate[i] * sexadjust, 1000);
      deadwomen = RESCALE(fem[i], deathrate, 1000000L);
      inew_births += RESCALE(fem[i] - deadwomen, brate[i], 1000000L);
    }
    inew_births -= RESCALE(inew_births, sptr->deathrate[stepcnt][0], 1000000L);

    /* Age the population */
    adj = IDIV(drate[COHORTS-2] * sexadjust, 1000);
    fem[COHORTS-1] += fem[COHORTS-2] -
		      RESCALE(fem[COHORTS-2], drate[COHORTS-2] - adj, 1000000L);
    male[COHORTS-1] += male[COHORTS-2] -
		    RESCALE(male[COHORTS-2], drate[COHORTS-2] + adj, 1000000L);
    adj = IDIV(drate[COHORTS-1] * sexadjust, 1000);
    fem[COHORTS-1] -= RESCALE(fem[COHORTS-1], drate[COHORTS-1] - adj, 1000000L);
    male[COHORTS-1] -= RESCALE(male[COHORTS-1], drate[COHORTS-1] + adj, 1000000L);
    if (fem[COHORTS-1] < 0) fem[COHORTS-1] = 0;
    if (male[COHORTS-1] < 0) male[COHORTS-1] = 0;
    for (i=COHORTS-2; i>0; i--)
    {
      adj = IDIV(drate[i-1] * sexadjust, 1000);
      fem[i] = fem[i-1] - RESCALE(fem[i-1], drate[i-1] - adj, 1000000L);
      male[i] = male[i-1] - RESCALE(male[i-1], drate[i-1] + adj, 1000000L);
      if (fem[i] < 0) fem[i] = 0;
      if (male[i] < 0) male[i] = 0;
    }
    adj = IDIV(drate[0] * sexadjust, 1000);
    fem[0] = RESCALE(inew_births, FEMALEPERCENT, 1000);
    male[0] = inew_births - fem[0];
    fem[0] -= RESCALE(fem[0], drate[0] - adj, 1000000L);
    male[0] -= RESCALE(male[0], drate[0] + adj, 1000000L);
    if (fem[0] < 0) fem[0] = 0;
    if (male[0] < 0) male[0] = 0;

    /* Update the infant mort, life expect, TFR and migration if required */
    calc_sim_tfr_variant(sptr);
    calc_sim_life_variant(sptr);
    calc_sim_mig_variant(sptr);

    /* Add in (or remove) migrants. */
    if (mig != 0)
      for (i=0; i<COHORTS; i++)
      {
	male[i] += RESCALE(mig, Mmig[i/5], 500);
	fem[i] += RESCALE(mig, Fmig[i/5], 500);
        if (male[i] < 0) male[i] = 0;
        if (fem[i] < 0) fem[i] = 0;
      }

    /* Finally, make other required updates */
    sptr->year++;
  }

  /* final patchups */
  reconstruct_pop(sptr, male, fem);
  for (sptr->totpop[stepcnt]=0, i=0; i<17; i++)
    sptr->totpop[stepcnt] += sptr->male_pop[stepcnt][i] +
                             sptr->fem_pop[stepcnt][i];
  if (sptr->totpop[stepcnt] > Display[Cdisp].maxpop)
    Display[Cdisp].maxpop = sptr->totpop[stepcnt];
  if (sptr->tfr_targ_year[stepcnt] < sptr->year)
  {
    sptr->tfr_targ_year[stepcnt] = sptr->year;
    sptr->tfr_targ_val[stepcnt] = get_TFR(sptr->birthrate[stepcnt]);
  }
  if (sptr->lifex_targ_year[stepcnt] < sptr->year)
  {
    sptr->lifex_targ_year[stepcnt] = sptr->year;
    sptr->lifex_targ_val[stepcnt] = get_life_expect(sptr->deathrate[stepcnt]);
  }
  if (sptr->infmort_targ_year[stepcnt] < sptr->year)
  {
    sptr->infmort_targ_year[stepcnt] = sptr->year;
    sptr->infmort_targ_val[stepcnt] = sptr->deathrate[stepcnt][0];
  }
  if (sptr->mig_targ_year[stepcnt] < sptr->year)
  {
    sptr->mig_targ_year[stepcnt] = sptr->year;
    sptr->mig_targ_val[stepcnt] = sptr->mig[stepcnt];
  }
}
